package tiny;

import java.util.List;

abstract class Visitor {
	void visit(Program tp) {
		for(Command cmd : tp.body)
			cmd.accept(this);
	}
	
	void visit(VarCommand tvc) {
	}
	
	void visit(IfCommand tif) {
		tif.cond.accept(this);
		for(Command cmd : tif.cmd_then)
			cmd.accept(this);
		if(tif.cmd_else != null) 
			for(Command cmd : tif.cmd_else)
				cmd.accept(this);
	}
	
	void visit(RepeatCommand tr) {
		for(Command cmd : tr.cmd_body)
			cmd.accept(this);
		tr.cond.accept(this);
	}
	
	void visit(AssignCommand ta) {
		ta.rval.accept(this);
	}
	
	void visit(ReadCommand tr) {
	}
	
	void visit(WriteCommand tw) {
		tw.exp.accept(this);
	}
	
	void visit(NumeralExpression tn) {
	}

	void visit(VarExpression tn) {
	}
	
	void visit(UnmExpression tu) {
		tu.exp.accept(this);
	}
	
	void visit(LtExpression tb) {
		tb.left.accept(this);
		tb.right.accept(this);
	}

	void visit(EqExpression tb) {
		tb.left.accept(this);
		tb.right.accept(this);
	}

	void visit(AddExpression tb) {
		tb.left.accept(this);
		tb.right.accept(this);
	}

	void visit(SubExpression tb) {
		tb.left.accept(this);
		tb.right.accept(this);
	}
	
	void visit(MulExpression tb) {
		tb.left.accept(this);
		tb.right.accept(this);
	}

	void visit(DivExpression tb) {
		tb.left.accept(this);
		tb.right.accept(this);
	}
}

abstract class Node {
	abstract void accept(Visitor v);
}

abstract class Command extends Node {
	void codegen(CodeGen ctx) {
		throw new RuntimeException("no implementado");
	}
}

abstract class Expression extends Node {
	public int type;
	
	public static int INT = 1;
	public static int BOOL = 2;

	void codegen(CodeGen ctx, Integer target) {
		throw new RuntimeException("no implementado");
	}
}

class Program extends Command {
	public List<Command> body;
	public SymbolTable st_body;
	
	Program(List<Command> body) {
		this.body = body;
	}

	void accept(Visitor tv) {
		tv.visit(this);
	}

	void codegen(CodeGen ctx) {
		ctx.st = st_body;
		ctx.emit("function main:");
		for(Command cmd : body)
			cmd.codegen(ctx);
		ctx.emit("RETURN", "R0","1");
	}

}

class VarCommand extends Command {
	String name;
	
	VarCommand(String name) {
		this.name = name;
	}
	
	void accept(Visitor tv) {
		tv.visit(this);
	}

	void codegen(CodeGen ctx) {
		ctx.emit("LOADK", "R" + ctx.st.get(name), "0");
	}
}

class IfCommand extends Command {
	public Expression cond;
	public List<Command> cmd_then;
	public List<Command> cmd_else;
	public SymbolTable st_then;
	public SymbolTable st_else;
	
	IfCommand(Expression cond, List<Command> cmd_then, List<Command> cmd_else) {
		this.cond = cond;
		this.cmd_then = cmd_then;
		this.cmd_else = cmd_else;
	}

	void accept(Visitor tv) {
		tv.visit(this);
	}
}

class RepeatCommand extends Command {
	public Expression cond;
	public List<Command> cmd_body;
	public SymbolTable st_body;
	
	RepeatCommand(List<Command> cmd_body, Expression cond) {
		this.cond = cond;
		this.cmd_body = cmd_body;
	}

	void accept(Visitor tv) {
		tv.visit(this);
	}
}

class AssignCommand extends Command {
	public String lval;
	public Expression rval;
	
	AssignCommand(String lval, Expression rval) {
		this.lval = lval;
		this.rval = rval;
	}
	
	void accept(Visitor tv) {
		tv.visit(this);
	}

	void codegen(CodeGen ctx) {
		rval.codegen(ctx, ctx.st.get(lval));
	}
}

class ReadCommand extends Command {
	public String var;
	
	ReadCommand(String var) {
		this.var = var;
	}

	void accept(Visitor tv) {
		tv.visit(this);
	}
	
	void codegen(CodeGen ctx) {
		throw new RuntimeException("no implementado");
	}
}

class WriteCommand extends Command {
	public Expression exp;
	
	WriteCommand(Expression exp) {
		this.exp = exp;
	}

	void accept(Visitor tv) {
		tv.visit(this);
	}

	void codegen(CodeGen ctx) {
		int print = ctx.temp++;
		int pexp = ctx.temp++;
		ctx.emit("GETGLOBAL", "R" + print, "print");
		exp.codegen(ctx, pexp);
		ctx.emit("CALL", "R" + print, "2", "1");
		ctx.temp = ctx.temp - 2;
	}
}

class NumeralExpression extends Expression {
	public int val;
	
	NumeralExpression(int val) {
		this.val = val;
	}

	void accept(Visitor tv) {
		tv.visit(this);
	}
	
	void codegen(CodeGen ctx, Integer target) {
		ctx.emit("LOADK", "R" + target, String.valueOf(val));
	}
}

class VarExpression extends Expression {
	public String name;
	
	VarExpression(String name) {
		this.name = name;
	}

	void accept(Visitor tv) {
		tv.visit(this);
	}

	void codegen(CodeGen ctx, Integer target) {
		int var = ctx.st.get(name);
		if(var != target)
			ctx.emit("MOVE", "R" + target, " R" + var);
	}
}

class UnmExpression extends Expression {
	public Expression exp;
	
	UnmExpression(Expression exp) {
		this.exp = exp;
	}

	void accept(Visitor tv) {
		tv.visit(this);
	}
	
	void codegen(CodeGen ctx, Integer target) {
		exp.codegen(ctx, target);
		ctx.emit("UNM", "R" + target, "R" + target);
	}
}

abstract class BinopExpression extends Expression {
	public Expression left;
	public Expression right;
}

class LtExpression extends BinopExpression {
	LtExpression(Expression left, Expression right) {
		this.left = left;
		this.right = right;
	}
	
	void accept(Visitor tv) {
		tv.visit(this);
	}
}

class EqExpression extends BinopExpression {
	EqExpression(Expression left, Expression right) {
		this.left = left;
		this.right = right;
	}

	void accept(Visitor tv) {
		tv.visit(this);
	}
}

class AddExpression extends BinopExpression {
	AddExpression(Expression left, Expression right) {
		this.left = left;
		this.right = right;
	}

	void accept(Visitor tv) {
		tv.visit(this);
	}

	void codegen(CodeGen ctx, Integer target) {
		left.codegen(ctx, target);
		int r = ctx.temp++;
		right.codegen(ctx, r);
		ctx.emit("ADD", "R" + target, "R" +
					target, "R" + r);
		ctx.temp--;
	}
}

class SubExpression extends BinopExpression {
	SubExpression(Expression left, Expression right) {
		this.left = left;
		this.right = right;
	}

	void accept(Visitor tv) {
		tv.visit(this);
	}
}

class MulExpression extends BinopExpression {
	MulExpression(Expression left, Expression right) {
		this.left = left;
		this.right = right;
	}

	void accept(Visitor tv) {
		tv.visit(this);
	}
}

class DivExpression extends BinopExpression {
	DivExpression(Expression left, Expression right) {
		this.left = left;
		this.right = right;
	}

	void accept(Visitor tv) {
		tv.visit(this);
	}
}
